#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base
import brewery.ds as ds

# data_sources = {
#     "csv": {"class": CSVDataSource},
#     "xls": {"class": XLSDataSource},
#     "yamldir": {"class": YamlDirectoryDataSource},
#     "mongodb": {"class": MongoDBDataSource},
#     "sql": {"class": SQLDataSource}
# }

class RowListSourceNode(base.SourceNode):
    """Source node that feeds rows (list/tuple of values) from a list (or any other iterable)
    object."""

    node_info = {
        "label" : "Row List Source",
        "description" : "Provide list of lists or tuples as data source.",
        "protected": True,
        "attributes" : [
            {
                 "name": "list",
                 "description": "List of rows represented as lists or tuples."
            },
            {
                 "name": "fields",
                 "description": "Fields in the list."
            }
        ]
    }
    def __init__(self, a_list = None, fields = None):
        if a_list:
            self.list = a_list
        else:
            self.list = []
        self.fields = fields
        
    @property
    def output_fields(self):
        if not self.fields:
            raise ValueError("Fields are not initialized")
        return self.fields

    def run(self):
        for row in self.list:
            self.put(row)

class RecordListSourceNode(base.SourceNode):
    """Source node that feeds records (dictionary objects) from a list (or any other iterable)
    object."""

    node_info = {
        "label" : "Record List Source",
        "description" : "Provide list of dict objects as data source.",
        "protected": True,
        "attributes" : [
            {
                 "name": "a_list",
                 "description": "List of records represented as dictionaries."
            },
            {
                 "name": "fields",
                 "description": "Fields in the list."
            }
        ]
    }

    def __init__(self, a_list = None, fields = None):
        if a_list:
            self.list = a_list
        else:
            self.list = []
        self.fields = fields

    @property
    def output_fields(self):
        if not self.fields:
            raise ValueError("Fields are not initialized")
        return self.fields

    def run(self):
        for record in self.list:
            self.put(record)
            
class StreamSourceNode(base.SourceNode):
    """Generic data stream source. Wraps a :mod:`brewery.ds` data source and feeds data to the 
    output.

    The source data stream should configure fields on initialize().

    Note that this node is only for programatically created processing streams. Not useable
    in visual, web or other stream modelling tools.
    """
    
    node_info = {
        "label" : "Data Stream Source",
        "icon": "row_list_source_node",
        "description" : "Generic data stream data source node.",
        "protected": True,
        "attributes" : [
            {
                 "name": "stream",
                 "description": "Data stream object."
            }
        ]
    }

    def __init__(self, stream):
        super(StreamSourceNode, self).__init__()
        self.stream = stream

    def initialize(self):
        # if self.stream_type not in data_sources:
        #     raise ValueError("No data source of type '%s'" % stream_type)
        # stream_info = data_sources[self.stream_type]
        # if "class" not in stream_info:
        #     raise ValueError("No stream class specified for data source of type '%s'" % stream_type)

        # self.stream = stream_class(**kwargs)
        # self.stream.fields = 
        self.stream.initialize()

    @property
    def output_fields(self):
        return self.stream.fields
        
    def run(self):
        for row in self.stream.rows():
            self.put(row)
        
    def finalize(self):
        self.stream.finalize()

class CSVSourceNode(base.SourceNode):
    """Source node that reads comma separated file from a filesystem or a remote URL.

    It is recommended to configure node fields before running. If you do not do so, fields are
    read from the file header if specified by `read_header` flag. Field storage types are set to
    `string` and analytical type is set to `typeless`.

    """
    node_info = {
        "label" : "CSV Source",
        "icon": "csv_file_source_node",
        "description" : "Read data from a comma separated values (CSV) file.",
        "attributes" : [
            {
                 "name": "resource",
                 "description": "File name or URL containing comma separated values",
            },
            {
                 "name": "fields",
                 "description": "fields contained in the file",
                 "type": "fields"
            },
            {
                 "name": "read_header",
                 "description": "flag determining whether first line contains header or not",
                 "type": "flag",
                 "default": "True"
            },
            {
                 "name": "skip_rows",
                 "description": "number of rows to be skipped",
                 "type": "flag"
            },
            {
                 "name": "encoding",
                 "description": "resource data encoding, by default no conversion is performed"
            },
            {
                 "name": "delimiter",
                 "description": "record delimiter character, default is comma ','"
            },
            {
                 "name": "quotechar",
                 "description": "character used for quoting string values, default is double quote"
            }
        ]
    }
    def __init__(self, resource = None, *args, **kwargs):
        super(CSVSourceNode, self).__init__()
        self.resource = resource
        self.args = args
        self.kwargs = kwargs
        self.stream = None
        self.fields = None
        self._output_fields = None
        
    @property
    def output_fields(self):
        if not self.stream:
            raise ValueError("Stream is not initialized")

        if not self._output_fields:
            raise ValueError("Fields are not initialized")

        return self._output_fields

    def initialize(self):
        self.stream = ds.CSVDataSource(self.resource, *self.args, **self.kwargs)
        
        if self.fields:
            self.stream.fields = self.fields
        
        self.stream.initialize()
        
        # FIXME: this is experimental form of usage
        self._output_fields = self.stream.fields.copy()
        self._output_fields.retype(self._retype_dictionary)

    def run(self):
        for row in self.stream.rows():
            self.put(row)
            
    def finalize(self):
        self.stream.finalize()

class XLSSourceNode(base.SourceNode):
    """Source node that reads Excel XLS files.

    It is recommended to configure node fields before running. If you do not do so, fields are
    read from the file header if specified by `read_header` flag. Field storage types are set to
    `string` and analytical type is set to `typeless`.

    """
    node_info = {
        "label" : "XLS Source",
        "icon": "xls_file_source_node",
        "description" : "Read data from an Excel (XLS) spreadsheet file.",
        "attributes" : [
            {
                 "name": "resource",
                 "description": "File name or URL containing comma separated values"
            },
            {
                 "name": "fields",
                 "description": "fields contained in the file",
            },
            {
                 "name": "sheet",
                 "description": "Sheet index number (as int) or sheet name (as string)"
            },
            {
                 "name": "read_header",
                 "description": "flag determining whether first line contains header or not",
                 "default": "True"
            },
            {
                 "name": "skip_rows",
                 "description": "number of rows to be skipped"
            },
            {
                 "name": "encoding",
                 "description": "resource data encoding, by default no conversion is performed"
            }
        ]
    }
    def __init__(self, *args, **kwargs):
        super(XLSSourceNode, self).__init__()
        self.args = args
        self.kwargs = kwargs
        self.stream = None
        self._fields = None

    @property
    def output_fields(self):
        if not self.stream:
            raise ValueError("Stream is not initialized")

        if not self.stream.fields:
            raise ValueError("Fields are not initialized")

        return self.stream.fields

    def __set_fields(self, fields):
        self._fields = fields
        if self.stream:
            self.stream.fields = fields

    def __get_fields(self):
        return self._fields

    fields = property(__get_fields, __set_fields)

    def initialize(self):
        self.stream = ds.XLSDataSource(*self.args, **self.kwargs)

        if self._fields:
            self.stream.fields = self._fields

        self.stream.initialize()
        self._fields = self.stream.fields

    def run(self):
        for row in self.stream.rows():
            self.put(row)

    def finalize(self):
        self.stream.finalize()


class YamlDirectorySourceNode(base.SourceNode):
    """Source node that reads data from a directory containing YAML files.
    
    The data source reads files from a directory and treats each file as single record. For example,
    following directory will contain 3 records::
    
        data/
            contract_0.yml
            contract_1.yml
            contract_2.yml
    
    Optionally one can specify a field where file name will be stored.
    """
    node_info = {
        "label" : "YAML Directory Source",
        "icon": "yaml_directory_source_node",
        "description" : "Read data from a directory containing YAML files",
        "protected": True,
        "attributes" : [
            {
                 "name": "path",
                 "description": "Path to a directory"
            },
            {
                 "name": "extension",
                 "description": "file extension to look for, default is yml. If none is given, "
                                "then all regular files in the directory are read.",
                 "default": "yml"
            },
            {
                 "name": "filename_field",
                 "description": "name of a new field that will contain file name",
                 "default": "True"
            }
        ]
    }
    def __init__(self, *args, **kwargs):
        super(YamlDirectorySourceNode, self).__init__()
        self.kwargs = kwargs
        self.args = args
        self.stream = None
        self.fields = None

    @property
    def output_fields(self):
        if not self.stream:
            raise ValueError("Stream is not initialized")

        if not self.stream.fields:
            raise ValueError("Fields are not initialized")

        return self.stream.fields

    def initialize(self):
        self.stream = ds.YamlDirectoryDataSource(*self.args, **self.kwargs)

        self.stream.fields = self.fields
        self.stream.initialize()

    def run(self):
        for row in self.stream.rows():
            # logging.debug("putting yaml row. pipe status: %s" % self.outputs[0].stop_sending)
            self.put(row)

    def finalize(self):
        self.stream.finalize()

class GoogleSpreadsheetSourceNode(base.SourceNode):
    """Source node that reads Google Spreadsheet.

    You should provide either spreadsheet_key or spreadsheet_name, if more than one spreadsheet with
    given name are found, then the first in list returned by Google is used.
    
    For worksheet selection you should provide either worksheet_id or worksheet_name. If more than
    one worksheet with given name are found, then the first in list returned by Google is used. If
    no worksheet_id nor worksheet_name are provided, then first worksheet in the workbook is used.
    
    For details on query string syntax see the section on sq under
    http://code.google.com/apis/spreadsheets/reference.html#list_Parameters
    """
    node_info = {
        "label" : "Google Spreadsheet Source",
        "icon": "google_spreadsheet_source_node",
        "description" : "Read data from a Google Spreadsheet.",
        "attributes" : [
            {
                 "name": "spreadsheet_key",
                 "description": "The unique key for the spreadsheet"
            },
            {
                 "name": "spreadsheet_name",
                 "description": "The title of the spreadsheets",
            },
            {
                 "name": "worksheet_id",
                 "description": "ID of a worksheet"
            },
            {
                 "name": "worksheet_name",
                 "description": "name of a worksheet"
            },
            {
                 "name": "query_string",
                 "description": "optional query string for row selection"
            },
            {
                 "name": "username",
                 "description": "Google account user name"
            },
            {
                 "name": "password",
                 "description": "Google account password"
            }
        ]
    }
    def __init__(self, *args, **kwargs):
        super(GoogleSpreadsheetSourceNode, self).__init__()
        self.args = args
        self.kwargs = kwargs
        self.stream = None
        self._fields = None

    @property
    def output_fields(self):
        if not self.stream:
            raise ValueError("Stream is not initialized")

        if not self.stream.fields:
            raise ValueError("Fields are not initialized")

        return self.stream.fields

    def __getattr__(self, key):
        try:
            return getattr(self.stream, key)
        except AttributeError:
            return object.__getattr__(self, key)

    def __set_fields(self, fields):
        self._fields = fields
        if self.stream:
            self.stream.fields = fields

    def __get_fields(self):
        return self._fields

    fields = property(__get_fields, __set_fields)

    def initialize(self):
        self.stream = ds.GoogleSpreadsheetDataSource(*self.args, **self.kwargs)

        if self._fields:
            self.stream.fields = self._fields

        self.stream.initialize()
        self._fields = self.stream.fields

    def run(self):
        for row in self.stream.rows():
            self.put(row)

    def finalize(self):
        self.stream.finalize()


class SQLSourceNode(base.SourceNode):
    """Source node that reads from a sql table.
    """
    node_info = {
        "label" : "SQL Source",
        "icon": "sql_source_node",
        "description" : "Read data from a sql table.",
        "attributes" : [
            {
                 "name": "uri",
                 "description": "SQLAlchemy URL"
            },
            {
                 "name": "table",
                 "description": "table name",
            },
        ]
    }
    def __init__(self, *args, **kwargs):
        super(SQLSourceNode, self).__init__()
        self.args = args
        self.kwargs = kwargs
        self.stream = None
        self._fields = None
        
    @property
    def output_fields(self):
        if not self.stream:
            raise ValueError("Stream is not initialized")

        if not self.stream.fields:
            raise ValueError("Fields are not initialized")

        return self.stream.fields

    def __set_fields(self, fields):
        self._fields = fields
        if self.stream:
            self.stream.fields = fields

    def __get_fields(self):
        return self._fields

    fields = property(__get_fields, __set_fields)

    def initialize(self):
        self.stream = ds.SQLDataSource(*self.args, **self.kwargs)
        self.stream.initialize()
        self._fields = self.stream.fields

    def run(self):
        for row in self.stream.rows():
            self.put(row)
            
    def finalize(self):
        self.stream.finalize()

class GeneratorFunctionSourceNode(base.SourceNode):
    """Source node uses a callable to generate records."""

    node_info = {
        "label" : "Callable Generator Source",
        "description" : "Uses a callable as record generator",
        "protected": True,
        "attributes" : [
            {
                 "name": "function",
                 "description": "Function (or any callable)"
            },
            {
                 "name": "fields",
                 "description": "Fields the function generates"
            },
            {
                 "name": "args",
                 "description": "Function arguments"
            },
            {
                 "name": "kwargs",
                 "description": "Function key-value arguments"
            }
        ]
    }

    def __init__(self, function=None, fields=None, *args, **kwargs):
        super(GeneratorFunctionSourceNode, self).__init__()

        self.function = function
        self.fields = fields
        self.args = args
        self.kwargs = kwargs

    @property
    def output_fields(self):
        if not self.fields:
            raise ValueError("Fields are not initialized")
        return self.fields

    def run(self):
        for row in self.function(*self.args, **self.kwargs):
            self.put(row)




# -*- coding: utf-8 -*-

import logging
import threading
import traceback
import sys
from brewery.utils import get_logger
from brewery.nodes import *
from brewery.common import *

__all__ = [
    "Stream",
    "Pipe",
    "stream_from_dict",
    "create_builder"
]

JOIN_TIMEOUT = None

def stream_from_dict(desc):
    """Create a stream from dictionary `desc`."""
    stream = Stream()
    stream.update(desc)
    return stream

class SimpleDataPipe(object):
    """Dummy pipe for testing nodes"""
    def __init__(self):
        self.buffer = []
        self.fields = None
        self._closed = False

    def closed(self):
        return self._closed

    def rows(self):
        return self.buffer

    def records(self):
        """Get data objects from pipe as records (dict objects). This is convenience method with
        performance costs. Nodes are recommended to process rows instead."""
        if not self.fields:
            raise Exception("Can not provide records: fields for pipe are not initialized.")
        fields = self.fields.names()
        for row in self.rows():
            yield dict(zip(fields, row))

    def put_record(self, record):
        """Convenience method that will transform record into a row based on pipe fields."""
        row = [record.get(field) for field in self.fields.names()]

        self.put(row)

    def put(self, obj):
        self.buffer.append(obj)

    def done_receiving(self):
        self._closed = True
        pass

    def done_sending(self):
        pass

    def empty(self):
        self.buffer = []

class Pipe(SimpleDataPipe):
    """Data pipe:
    Contains buffer for data that should be thransferred to another node.
    Data are being sent t other node when the buffer is full. Pipe is one-directional where
    one thread is sending data to another thread. There is only one backward signalling: closing 
    the pipe from remote object.


    """

    def __init__(self, buffer_size=1000):
        """Creates uni-drectional data pipe for passing data between two threads in batches of size
        `buffer_size`.

        If receiving node is finished with source data and does not want anything any more, it
        should send ``done_receiving()`` to the pipe. In most cases, stream runner will send
        ``done_receiving()`` to all input pipes when node's ``run()`` method is finished.

        If sending node is finished, it should send ``done_sending()`` to the pipe, however this
        is not necessary in most cases, as the method for running stream flushes outputs
        automatically on when node ``run()`` method is finished.
        """

        super(Pipe, self).__init__()
        self.buffer_size = buffer_size

        # Should it be deque or array?
        self.staging_buffer = []
        self._ready_buffer = None

        self._done_sending = False
        self._done_receiving = False
        self._closed = False

        # Taken from Python Queue implementation:

        # mutex must beheld whenever the queue is mutating.  All methods
        # that acquire mutex must release it before returning.  mutex
        # is shared between the three conditions, so acquiring and
        # releasing the conditions also acquires and releases mutex.
        self.mutex = threading.Lock()
        # Notify not_empty whenever an item is added to the queue; a
        # thread waiting to get is notified then.
        self.not_empty = threading.Condition(self.mutex)
        # Notify not_full whenever an item is removed from the queue;
        # a thread waiting to put is notified then.
        self.not_full = threading.Condition(self.mutex)

    def is_full(self):
        return len(self.staging_buffer) >= self.buffer_size

    def is_consumed(self):
        return self._ready_buffer is None

    def put(self, obj):
        """Put data object into the pipe buffer. When buffer is full it is enqueued and receiving node
        can get all buffered data objects.

        Puttin object into pipe is not thread safe. Only one thread sohuld write to the pipe.
        """
        self.staging_buffer.append(obj)

        if self.is_full():
            self._flush()
    def _note(self, note):
        # print note
        pass

    def _flush(self, close=False):
        self._note("P flushing: close? %s closed? %s" % (close, self._closed))
        self._note("P _nf acq?")
        self.not_full.acquire()
        if self._closed:
            self._note("P _not_full rel!")
            self.not_full.release()
            return
        elif len(self.staging_buffer) == 0:
            try:
                self._closed = close
                self.not_empty.notify()
            finally:
                self._note("P _not_full rel!")
                self.not_full.release()
            return

        try:
            self._note("P _not_full wait ...")
            while not self.is_consumed() and not self._closed:
                self.not_full.wait()
            self._note("P _not_full got <")
            if not self._closed:
                self._ready_buffer = self.staging_buffer
                self.staging_buffer = []
                self._closed = close
                self._note("P _not_empty notify >")
                self.not_empty.notify()

        finally:
            self._note("P _not_full rel!")
            self.not_full.release()

    def rows(self):
        """Get data object from pipe. If there is no buffer ready, wait until source object sends
        some data."""

        done_sending = False
        while(not done_sending):
            self._note("C _not_empty acq?")
            self.not_empty.acquire()
            try:
                self._note("C _not_empty wait ...")
                while not self._ready_buffer and not self._closed:
                    self.not_empty.wait()
                self._note("C _not_empty got <")

                if self._ready_buffer:
                    rows = self._ready_buffer
                    self._ready_buffer = None
                    self._note("C _not_full notify >")
                    self.not_full.notify()

                    for row in rows:
                        yield row
                else:
                    self._note("C no buffer")


                done_sending = self._closed
            finally:
                self._note("_not_empty rel!")
                self.not_empty.release()

    def closed(self):
        """Return ``True`` if pipe is closed - not sending or not receiving data any more."""
        return self._closed

    def done_sending(self):
        """Close pipe from sender side"""
        self._flush(True)

    def done_receiving(self):
        """Close pipe from either side"""
        self._note("C not_empty acq? r")
        self.not_empty.acquire()
        self._note("C closing")
        self._closed = True
        self._note("C notif close")
        self.not_full.notify()
        self.not_empty.release()

        self._note("C not_empty rel! r")

class Stream(object):
    """Data processing stream"""
    def __init__(self, nodes=None, connections=None):
        """Creates a data stream.

        :Parameters:
            * `nodes` - dictionary with keys as node names and values as nodes
            * `connections` - list of two-item tuples. Each tuple contains source and target node
              or source and target node name.
            * `stream` - another stream or 
        """
        super(Stream, self).__init__()
        self.nodes = []
        self.node_dict = {}
        self.connections = set()

        self.logger = get_logger()

        if nodes:
            try:
                for name, node in nodes.items():
                    self.add(node, name)
            except:
                raise StreamError("Nodes should be a dictionary, is %s" % type(nodes))

        if connections:
            for connection in connections:
                self.connect(connection[0], connection[1])

        self.exceptions = []

    def node(self, node):
        """Returns a node in the stream or node with name. This method is used for coalescing in
        other methods, where you can pass either node name or node object.
        
        Raises KeyError when node does not exist.

        :Parameters:
            * `node` - node object or node name
        """

        if type(node) == str or type(node) == unicode:
            return self.node_dict[node]
        else:
            return node

    def add(self, node, name=None):
        """Add a `node` into the stream. Does not allow to add named node if node with given name
        already exists. """
        if name:
            if name in self.node_dict:
                raise KeyError("Node with name %s already exists" % name)
            self.node_dict[name] = node

        if node not in self.nodes:
            self.nodes.append(node)

    def set_node_name(self, node, name):
        """Sets a name for `node`. If `name` is ``None`` then node name will be removed.
        
        Raises an exception if the `node` is not part of the stream or there is already node with
        same name.
        """
        if node not in self.nodes:
            raise KeyError("Node %s does not belong to stream" % node)

        if name:
            if name in self.node_dict:
                raise KeyError("Node with name %s already exists" % name)

            self.node_dict[name] = node
        else:
            del self.node_dict[name]

    def remove(self, node):
        """Remove a `node` from the stream. Also all connections will be removed."""

        node = self.node(node)

        self.nodes.remove(node)
        for (name, current_node) in self.node_dict.items():
            if current_node == node:
                del self.node_dict[name]

        to_be_removed = set()
        for connection in self.connections:
            if connection[0] == node or connection[1] == node:
                to_be_removed.add(connection)

        for connection in to_be_removed:
            self.connections.remove(connection)

    def connect(self, source, target):
        """Connects source node and target node. Nodes can be provided as objects or names."""

        source_node = self.node(source)
        target_node = self.node(target)
        self.connections.add((source_node, target_node))

    def remove_connection(self, source, target):
        """Remove connection between source and target nodes, if exists."""
        source_node = self.node(source)
        target_node = self.node(target)

        self.connections.discard((source_node, target_node))

    def sorted_nodes(self):
        """
        Return topologically sorted nodes.
        
        Algorithm::
        
            L = Empty list that will contain the sorted elements
            S = Set of all nodes with no incoming edges
            while S is non-empty do
                remove a node n from S
                insert n into L
                for each node m with an edge e from n to m do
                    remove edge e from the graph
                    if m has no other incoming edges then
                        insert m into S
            if graph has edges then
                raise exception: graph has at least one cycle
            else 
                return proposed topologically sorted order: L
        """
        def is_source(node, connections):
            for connection in connections:
                if node == connection[1]:
                    return False
            return True

        def source_connections(node, connections):
            conns = set()
            for connection in connections:
                if node == connection[0]:
                    conns.add(connection)
            return conns

        nodes = set(self.nodes)
        connections = self.connections.copy()
        sorted_nodes = []
        source_nodes = set()

        # Find source nodes:
        for node in nodes:
            if is_source(node, connections):
                source_nodes.add(node)

        # while S is non-empty do
        while source_nodes:
            # remove a node n from S
            node = source_nodes.pop()
            # insert n into L
            sorted_nodes.append(node)

            # for each node m with an edge e from n to m do
            s_connections = source_connections(node, connections)
            for connection in s_connections:
                #     remove edge e from the graph
                m = connection[1]
                connections.remove(connection)
                #     if m has no other incoming edges then
                #         insert m into S
                if is_source(m, connections):
                    source_nodes.add(m)

        # if graph has edges then
        #     output error message (graph has at least one cycle)
        # else 
        #     output message (proposed topologically sorted order: L)

        if connections:
            raise Exception("Steram has at least one cycle")

        return sorted_nodes
        
    def fork(self, node=None):
        """Creates a construction fork of the stream. Used for constructing streams in functional
        fashion. Example::
        
            stream = Stream()

            fork = stream.fork()
            fork.csv_source("fork.csv")
            fork.formatted_printer()

            stream.run()
        
        Fork responds to node names as functions. The function arguments are the same as node
        constructor (__init__ method) arguments. Each call will append new node to the fork and
        will connect the new node to the previous node in the fork.
        
        To configure current node you can use ``fork.node``, like::
        
            fork.csv_source("fork.csv")
            fork.node.read_header = True
            
        To set actual node name use ``set_name()``::

            fork.csv_source("fork.csv")
            fork.set_name("source")
            
            ...
            
            source_node = stream.node("source")
        
        To fork a fork, just call ``fork()``
        """

        return _StreamFork(self, self.node(node))


    def update(self, nodes = None, connections = None):
        """Adds nodes and connections specified in the dictionary. Dictionary might contain
        node names instead of real classes. You can use this method for creating stream
        from a dictionary that was created from a JSON file, for example.
        """

        node_dict = node_dictionary()

        # FIXME: use either node type identifier or fully initialized node, not
        #        node class (Warning: might break some existing code,
        #        depreciate it first
        
        nodes = nodes or {}
        connections = connections or []

        for (name, obj) in nodes.items():
            if isinstance(obj, Node):
                node_instance = obj
            elif isinstance(obj, type) and issubclass(obj, Node):
                self.logger.warn("Using classes in Stream.update is depreciated")
                node_instance = obj()
            else:
                if not "type" in obj:
                    raise Exception("Node dictionary has no 'type' key")
                node_type = obj["type"]

                if node_type in node_dict:
                    node_class = node_dict[node_type]
                    node_instance = node_class()

                    node_instance.configure(obj)
                else:
                    raise Exception("No node class of type '%s'" % node_type)

            self.add(node_instance, name)

        if connections:
            for connection in connections:
                self.connect(connection[0], connection[1])

    def configure(self, config = {}):
        """Configure node properties based on configuration. Only named nodes can be configured at the
        moment.

        `config` is a list of dictionaries with keys: ``node`` - node name, ``parameter`` - node parameter
        name, ``value`` - parameter value
        
        .. warning:
        
            This method might change to a list of dictionaries where one 
            dictionary will represent one node, keys will be attributes.
        
        """
        
        # FIXME: this is wrong, it should be a single dict per node (or not?)
        # List of attributes:
        #     * can reflect a form for configuring whole stream
        #     * can have attribute order regardless of their node ownership
        # List of nodes:
        #     * bundled attributes in single dictioary
        # FIXME: this is inconsistent with node configuration! node.config()
        
        configurations = {}

        # Collect configurations for each node
        
        for attribute in config:
            node_name = attribute["node"]
            attribute_name = attribute["attribute"]
            value = attribute.get("value")

            if not node_name in configurations:
                config = {}
                configurations[node_name] = config
            else:
                config = configurations[node_name]

            config[attribute_name] = value

        # Configure nodes

        for (node_name, config) in configurations.items():
            node = self.node(node_name)
            node.configure(config)
        
    def node_targets(self, node):
        """Return nodes that `node` passes data into."""
        node = self.node(node)
        nodes =[conn[1] for conn in self.connections if conn[0] == node]
        return nodes

    def node_sources(self, node):
        """Return nodes that provide data for `node`."""
        node = self.node(node)
        nodes =[conn[0] for conn in self.connections if conn[1] == node]
        return nodes

    def _initialize(self):
        """Initializes the data processing stream:
        
        * sorts nodes based on connection dependencies
        * creates pipes between nodes
        * initializes each node
        * initializes pipe fields
        
        """

        self.logger.info("initializing stream")
        self.logger.debug("sorting nodes")
        sorted_nodes = self.sorted_nodes()
        self.pipes = []

        self.logger.debug("flushing pipes")
        for node in sorted_nodes:
            node.inputs = []
            node.outputs = []

        # Create pipes and connect nodes
        for node in sorted_nodes:
            self.logger.debug("creating pipes for node %s" % node)

            targets = self.node_targets(node)
            for target in targets:
                self.logger.debug("  connecting with %s" % (target))
                pipe = Pipe()
                node.add_output(pipe)
                target.add_input(pipe)
                self.pipes.append(pipe)

        # Initialize fields
        for node in sorted_nodes:
            self.logger.debug("initializing node of type %s" % node.__class__)
            self.logger.debug("  node has %d inputs and %d outputs"
                                % (len(node.inputs), len(node.outputs)))
            node.initialize()

            # Ignore target nodes
            if isinstance(node, TargetNode):
                self.logger.debug("  node is target, ignoring creation of output pipes")
                continue

            fields = node.output_fields
            self.logger.debug("  node output fields: %s" % fields.names())
            for output_pipe in node.outputs:
                output_pipe.fields = fields

    def run(self):
        """Run all nodes in the stream.
        
        Each node is being wrapped and run in a separate thread.
        
        When an exception occurs, the stream is stopped and all catched exceptions are stored in
        attribute `exceptions`.
        
        """
        self._initialize()

        # FIXME: do better exception handling here: what if both will raise exception?
        try:
            self._run()
        finally:
            self._finalize()

    def _run(self):


        self.logger.info("running stream")

        threads = []
        sorted_nodes = self.sorted_nodes()

        self.logger.debug("launching threads")
        for node in sorted_nodes:
            self.logger.debug("launching thread for node %s" % node_label(node))
            thread = _StreamNodeThread(node)
            thread.start()
            threads.append((thread, node))

        self.exceptions = []
        for (thread, node) in threads:
            self.logger.debug("joining thread for %s" % node_label(node))
            while True:
                thread.join(JOIN_TIMEOUT)
                if thread.isAlive():
                    pass
                    # self.logger.debug("thread join timed out")
                else:
                    if thread.exception:
                        self._add_thread_exception(thread)
                    else:
                        self.logger.debug("thread joined")
                    break
                if self.exceptions:
                    self.logger.info("node exception occured, trying to kill threads")
                    self.kill_threads()

        if self.exceptions:
            self.logger.info("run finished with exception")
            # Raising only first exception found
            raise self.exceptions[0]
        else:
            self.logger.info("run finished sucessfully")

    def _add_thread_exception(self, thread):
        """Create a StreamRuntimeError exception object and fill attributes with all necessary
        values.
        """
        node = thread.node
        exception = StreamRuntimeError(node=node, exception=thread.exception)

        exception.traceback = thread.traceback
        exception.inputs = [pipe.fields for pipe in node.inputs]

        if not isinstance(node, TargetNode):
            try:
                exception.ouputs = node.output_fields
            except:
                pass

        node_info = node.__class__.__dict__.get("__node_info__")

        attrs = {}
        if node_info and "attributes" in node_info:
            for attribute in node_info["attributes"]:
                attr_name = attribute.get("name")
                if attr_name:
                    try:
                        value = getattr(node, attr_name)
                    except AttributeError:
                        value = "<attribute %s does not exist>" % attr_name
                    except Exception , e:
                        value = e
                    attrs[attr_name] = value

        exception.attributes = attrs

        self.exceptions.append(exception)


    def kill_threads(self):
        self.logger.info("killing threads")

    def _finalize(self):
        self.logger.info("finalizing nodes")

        # FIXME: encapsulate finalization in exception handler, collect exceptions
        for node in self.sorted_nodes():
            self.logger.debug("finalizing node %s" % node_label(node))
            node.finalize()
def node_label(node):
    """Debug label for a node: node identifier with python object id."""
    return "%s(%s)" % (node.identifier() or str(type(node)), id(node))
    
class _StreamNodeThread(threading.Thread):
    def __init__(self, node):
        """Creates a stream node thread.
        
        :Attributes:
            * `node`: a Node object
            * `exception`: attribute will contain exception if one occurs during run()
            * `traceback`: will contain traceback if exception occurs
        
        """
        super(_StreamNodeThread, self).__init__()
        self.node = node
        self.exception = None
        self.traceback = None
        self.logger = get_logger()

    def run(self):
        """Wrapper method for running a node"""
        
        label = node_label(self.node)
        self.logger.debug("%s: start" % label)
        try:
            self.node.run()
        except NodeFinished as e:
            self.logger.info("node %s finished" % (label))
        except Exception as e:
            tb = sys.exc_info()[2]
            self.traceback = tb

            self.logger.debug("node %s failed: %s" % (label, e.__class__.__name__), exc_info=sys.exc_info)
            self.exception = e

        # Flush pipes after node is finished
        self.logger.debug("%s: finished" % label)
        self.logger.debug("%s: flushing outputs" % label)
        for pipe in self.node.outputs:
            if not pipe.closed():
                pipe.done_sending()
        self.logger.debug("%s: flushed" % label)
        self.logger.debug("%s: stopping inputs" % label)
        for pipe in self.node.inputs:
            if not pipe.closed():
                pipe.done_sending()
        self.logger.debug("%s: stopped" % self)

class _StreamFork(object):
    """docstring for StreamFork"""
    def __init__(self, stream, node=None):
        """Creates a stream fork - class for building streams."""
        super(_StreamFork, self).__init__()
        self.stream = stream
        self.node = node

    def __iadd__(self, node):
        """Appends a node to the actual stream. The new node becomes actual node of the
        for."""

        self.stream.add(node)
        if self.node:
            self.stream.connect(self.node, node)
        self.node = node

        return self

    def set_name(self, name):
        """Sets name of current node."""
        self.stream.set_node_name(self.node, name)

    def fork(self):
        """Forks current fork. Returns a new fork with same actual node as the fork being
        forked."""
        fork = _StreamFork(self.stream, self.node)
        return fork

    def merge(self, obj, **kwargs):
        """Joins two streams using the MergeNode (please refer to the node documentaton
        for more information).
        
        `obj` is a fork or a node to be merged. `kwargs` are MergeNode configuration arguments,
        such as `joins`.
        
        """
        raise NotImplementedError
        # if type(obj) == StreamFork:
        #     node = obj.node
        # else:
        #     node = obj
        # 
        # self.stream.append(node)
        # 
        # merge = MergeNode(**kwargs)
        # self.stream.append(merge)
        # self.stream.connect()

    def append(self, obj):
        """Appends data from nodes using AppendNode"""
        raise NotImplementedError 

    def __getattr__(self, name):
        """Returns node class"""
        # FIXME: use create_node here
        
        class_dict = node_dictionary()

        node_class = class_dict[name]

        constructor = _StreamForkConstructor(self, node_class)
        return constructor

class _StreamForkConstructor(object):
    """Helper class to append new node."""
    def __init__(self, fork, node_class):
        self.fork = fork
        self.node_class = node_class

    def __call__(self, *args, **kwargs):
        node = self.node_class(*args, **kwargs)
        self.fork += node
        return self.fork

def create_builder():
    """Creates a stream builder for incremental stream building."""
    stream = Stream()
    return stream.fork()
    
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base

class SimpleHTMLDataTarget(base.DataTarget):
    def __init__(self, resource, html_header = True, html_footer = None, 
                 write_headers = True, table_attributes = None,
                ):
        """Creates a HTML data target with simple naive HTML generation. No package that generates
        document node tree is used, just plain string concatenation.
        
        :Attributes:
            * resource: target object - might be a filename or file-like object - you can stream
              HTML table data into existing opened file.
            * write_headers: create table headers, default: True. Field labels will be used,
              if field has no label, then fieln name will be used.
            * table_attributes: <table> node attributes such as ``class``, ``id``, ...
            * html_header: string to be used as HTML header. If set to ``None`` only <table> will
              be generated. If set to ``True`` then default header is used. Default is ``True``.
            * html_header: string to be used as HTML footer. Works in similar way as to html_header.

        Note: No HTML escaping is done. HTML tags in data might break the output.
        """

        self.resource = resource
        self.write_headers = write_headers
        self.table_attributes = table_attributes
        
        if html_header == True:
            self.html_header = """
            <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
            <html>
            <head>
            <meta http-equiv="Content-type" content="text/html; charset=utf-8" />
            </head>
            <body>
            """
        elif html_header and html_header != True:
            self.html_header = html_header
        else:
            self.html_header = ""
            
        if html_footer == True:
            self.html_footer = """
            </body>"""
        elif html_footer and html_footer != True:
            self.html_footer = html_footer
        else:
            self.html_footer = ""
        
    def initialize(self):
        self.handle, self.close_file = base.open_resource(self.resource, "w")

        if self.html_header:
            self.handle.write(self.html_header)

        attr_string = u""

        if self.table_attributes:
            for attr_value in self.table_attributes.items():
                attr_string += u' %s="%s"\n' % attr_value

        string = u"<table%s>\n" % attr_string
            
        if self.write_headers:
            string += u"<tr>"
            for field in self.fields:
                if field.label:
                    header = field.label
                else:
                    header = field.name
                
                string += u"  <th>%s</th>\n" % header

            string += u"</tr>\n"
            
        self.handle.write(string)
        
    def append(self, obj):
        if type(obj) == dict:
            row = []
            for field in self.field_names:
                row.append(obj.get(field))
        else:
            row = obj

        string = u"<tr>"
        for value in row:
            string += u"  <td>%s</td>\n" % value

        string += u"</tr>\n"
        self.handle.write(string.encode('utf-8'))

    def finalize(self):
        string = u"</table>"
        self.handle.write(string)

        if self.html_footer:
            self.handle.write(self.html_footer)
        
        if self.close_file:
            self.handle.close()
        


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Field statistics"""

import sets

class FieldStatistics(object):
    """Data quality statistics for a dataset field
    
    :Attributes:
        * `field`: name of a field for which statistics are being collected

        * `value_count`: number of records in which the field exist. In relationad database table this
          is equal to number of rows, in document based databse, such as MongoDB, it is number of
          documents that have a key present (being null or not)

        * `record_count`: total count of records in dataset. This should be set explicitly on
          finalisation. Seet :meth:`FieldStatistics.finalize`. In relational database this should be the
          same as `value_count`.

        * `value_ratio`: ratio of value count to record count, 1 for relational databases

        * `null_count`: number of records where field is null

        * `null_value_ratio`: ratio of records with nulls to total number of probed values =
          `null_value_ratio` / `value_count`

        * `null_record_ratio`: ratio of records with nulls to total number of records =
          `null_value_ratio` / `record_count`

        * `empty_string_count`: number of empty strings

        * `storage_types`: list of all encountered storage types (CSV, MongoDB, XLS might have different
          types within a field)

        * `unique_storage_type`: if there is only one storage type, then this is set to that type

        * `distict_values`: list of collected distinct values

        * `distinct_threshold`: number of distict values to collect, if count of distinct values is
          greather than threshold, collection is stopped and `distinct_overflow` will be set. Set to 0
          to get all values. Default is 10.
    """    
    def __init__(self, key = None, distinct_threshold = 10):
        self.field = key
        self.value_count = 0
        self.record_count = 0
        self.value_ratio = 0
        
        self.distinct_values = sets.Set()
        self.distinct_overflow = False
        self.storage_types = sets.Set()

        self.null_count = 0
        self.null_value_ratio = 0
        self.null_record_ratio = 0
        self.empty_string_count = 0
        
        self.distinct_threshold = distinct_threshold
        
        self.unique_storage_type = None
        
        self.probes = []
        
    def probe(self, value):
        """Probe the value:
        
        * increase found value count
        * identify storage type
        * probe for null and for empty string

        * probe distinct values: if their count is less than ``distinct_threshold``. If there are more
          distinct values than the ``distinct_threshold``, then distinct_overflow flag is set and list
          of distinct values will be empty
                    
        """
        
        storage_type = value.__class__
        self.storage_types.add(storage_type.__name__)

        self.value_count += 1
        
        # FIXME: check for existence in field.empty_values
        if value is None:
            self.null_count += 1

        if value == '':
            self.empty_string_count += 1

        self._probe_distinct(value)
        
        for probe in self.probes:
            probe.probe(value)

    def _probe_distinct(self, value):
        """"""
        if self.distinct_overflow:
            return

        # We are not testing lists, dictionaries and object IDs
        storage_type = value.__class__

        if not self.distinct_threshold or self.distinct_threshold == 0 or len(self.distinct_values) < self.distinct_threshold:
            try:
                self.distinct_values.add(value)
            except:
                # FIXME: Should somehow handle invalid values that can not be added
                pass
        else:
            self.distinct_overflow = True
            
    def finalize(self, record_count = None):
        """Compute final statistics.
        
        :Parameters:
            * `record_count`: final number of records in probed dataset.
                See :meth:`FieldStatistics` for more information.
        """
        if record_count:
            self.record_count = record_count
        else:
            self.record_count = self.value_count

        if self.record_count:
            self.value_ratio = float(self.value_count) / float(self.record_count)
            self.null_record_ratio = float(self.null_count) / float(self.record_count)

        if self.value_count:
            self.null_value_ratio = float(self.null_count) / float(self.value_count)

        if len(self.storage_types) == 1:
            self.unique_storage_type = list(self.storage_types)[0]

    def dict(self):
        """Return dictionary representation of receiver."""
        d = {}
        d["key"]= self.field
        d["value_count"]= self.value_count
        d["record_count"]= self.record_count
        d["value_ratio"]= self.value_ratio
        if self.distinct_overflow:
            d["distinct_overflow"] = self.distinct_overflow,
            d["distinct_values"] = []
        else:
            d["distinct_values"] = list(self.distinct_values)

        d["storage_types"]= list(self.storage_types)

        d["null_count"] = self.null_count
        d["null_value_ratio"] = self.null_value_ratio
        d["null_record_ratio"] = self.null_record_ratio
        d["empty_string_count"] = self.empty_string_count

        d["unique_storage_type"] = self.unique_storage_type
        
        return d
    
    def __repr__(self):
        return "FieldStatistics:%s" % (self.dict())




from setuptools import setup, find_packages

install_requires = []

try:
    # For Python >= 2.6
    import json
except ImportError:
    # For Python < 2.6 or people using a newer version of simplejson
    install_requires.append("simplejson")
    
setup(
    name = "brewery",
    version = '0.8.0',

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires=install_requires,

    packages=find_packages(exclude=['ez_setup']),
    # packages = ['brewery'],
	# package_dir = { 'brewery': 'brewery'},
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.txt', '*.rst'],
    },

    scripts=['bin/mongoaudit', 'bin/brewery'],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Database',
        'Topic :: Scientific/Engineering',
        'Topic :: Utilities'
    ],

    test_suite="brewery.tests",

    # metadata for upload to PyPI
    author="Stefan Urbanek",
    author_email="stefan.urbanek@gmail.com",
    description="Framework for analysing data and measuring quality of data using structured data streams",
    license="MIT",
    keywords="data analysis quality datamining",
    url="http://www.databrewery.org"

    # could also include long_description, download_url, classifiers, etc.
)



# -*- coding: utf-8 -*-

import unittest
import os
import json
import re

from test_data_source import *
from test_pipes import *
from test_nodes import *
from test_field_list import *
from test_node_stream import *
from test_data_quality import *
from test_sql_streams import *
from test_forks import *

test_cases = [FieldListCase,
              DataSourceUtilsTestCase,
              DataSourceTestCase,
              PipeTestCase,
              Pipe2TestCase,
              NodesTestCase,
              StreamBuildingTestCase,
              StreamInitializationTestCase,
              DataQualityTestCase,
              StreamConfigurationTestCase,
              SQLStreamsTestCase,
              ForksTestCase
                ]

def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    for test_class in test_cases:
        tests = loader.loadTestsFromTestCase(test_class)
        suite.addTests(tests)
    return suite





import unittest
import threading
import time
import brewery.ds as ds
import brewery.metadata

from sqlalchemy import Table, Column, Integer, String, Text
from sqlalchemy import create_engine, MetaData

class SQLStreamsTestCase(unittest.TestCase):
    def setUp(self):
        self.engine = create_engine("sqlite://")
        self.metadata = MetaData()
        
        self.fields = brewery.metadata.FieldList([
                            ("category", "string"),
                            ("category_label", "string"), 
                            ("subcategory", "string"), 
                            ("subcategory_label", "string"), 
                            ("line_item", "string"),
                            ("year", "integer"), 
                            ("amount", "integer")])
        self.example_row = ["cat", "Category", "scat", "Sub-category", "foo", 2012, 100]
        
    def test_table_fields(self):
        table = Table('users', self.metadata,
                    Column('id', Integer, primary_key=True),
                    Column('login', String(32)),
                    Column('full_name', String(255)),
                    Column('profile', Text)
                )
        
        self.metadata.create_all(self.engine)
        
        stream = ds.SQLDataSource(connection=self.engine, table=str(table))
        
        fields = stream.fields
        
        self.assertEqual(4, len(fields))
        
    def test_target_no_existing_table(self):
        stream = ds.SQLDataTarget(connection=self.engine, table="test")
        self.assertRaises(Exception, stream.initialize)

    def test_target_create_table(self):
        stream = ds.SQLDataTarget(connection=self.engine, table="test", create=True)
        # Should raise an exception, because no fields are specified
        self.assertRaises(Exception, stream.initialize)

        stream.fields = self.fields
        stream.initialize()

        cnames = [str(c) for c in stream.table.columns]
        fnames = ["test."+f.name for f in self.fields]
        self.assertEqual(fnames, cnames)

        stream.finalize()
        
    def test_target_replace_table(self):
        table = Table('test', self.metadata,
                    Column('id', Integer, primary_key=True),
                    Column('login', String(32)),
                    Column('full_name', String(255)),
                    Column('profile', Text)
                )
        
        self.metadata.create_all(self.engine)

        stream = ds.SQLDataTarget(connection=self.engine, table="test", 
                                    create=True, replace = False)
        
        stream.fields = self.fields
        self.assertRaises(Exception, stream.initialize)

        stream = ds.SQLDataTarget(connection=self.engine, table="test", 
                                    create=True, replace = True)
        stream.fields = self.fields
        stream.initialize()
        cnames = [str(c) for c in stream.table.columns]
        fnames = ["test."+f.name for f in self.fields]
        self.assertEqual(fnames, cnames)
        stream.finalize()
        
    def test_target_concrete_type_map(self):
        ctm = {"string": String(123)}
        stream = ds.SQLDataTarget(connection=self.engine, table="test",
                                  create=True,
                                  fields=self.fields,
                                  concrete_type_map=ctm)
        stream.initialize()

        c = stream.table.c["line_item"]

        self.assertEqual(123, c.type.length)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base
import datetime
from brewery.metadata import FieldList

try:
    import xlrd
except:
    from brewery.utils import MissingPackage
    xlrd = MissingPackage("xlrd", "Reading MS Excel XLS Files", "http://pypi.python.org/pypi/xlrd")

class XLSDataSource(base.DataSource):
    """Reading Microsoft Excel XLS Files

    Requires the xlrd package (see pypi).

    Based on the OKFN Swiss library.
    """
    def __init__(self, resource, sheet=None, encoding=None, skip_rows=None, read_header=True):
        """Creates a XLS spreadsheet data source stream.
        
        :Attributes:
            * resource: file name, URL or file-like object
            * sheet: sheet index number (as int) or sheet name (as str)
            * read_header: flag determining whether first line contains header or not. 
                ``True`` by default.
        """
        self.resource = resource
        self.sheet_reference = sheet
        self.read_header = read_header
        self.header_row = 0
        self.skip_rows = skip_rows
        self._fields = None
        self.close_file = True
        self.encoding = encoding
        self.fields = None

    def initialize(self):
        """Initialize XLS source stream:
        """

        self.file, self.close_file = base.open_resource(self.resource)

        self.workbook = xlrd.open_workbook(file_contents=self.file.read(),
                                           encoding_override=self.encoding)

        if not self.sheet_reference:
            self.sheet_reference = 0

        if type(self.sheet_reference) == int:
            self.sheet = self.workbook.sheet_by_index(self.sheet_reference)
        else:
            self.sheet = self.workbook.sheet_by_name(self.sheet_reference)

        self.row_count = self.sheet.nrows

        self.read_fields()

    def finalize(self):
        if self.file and self.close_file:
            self.file.close()

    def rows(self):
        if not self.sheet:
            raise RuntimeError("XLS Stream is not initialized - there is no sheet")
        if not self.fields:
            raise RuntimeError("Fields are not initialized")
        return XLSRowIterator(self.workbook, self.sheet, self.skip_rows)

    def records(self):
        fields = self.field_names
        for row in self.rows():
            yield dict(zip(fields, row))

    def read_fields(self):
        # FIXME: be more sophisticated and read field types from next row
        if self.read_header:
            row = self.sheet.row_values(self.header_row)
            self.fields = FieldList(row)
            self.skip_rows = self.header_row + 1

class XLSRowIterator(object):
    """
    Iterator that reads XLS spreadsheet
    """
    def __init__(self, workbook, sheet, row_offset=0):
        self.workbook = workbook
        self.sheet = sheet
        self.row_count = sheet.nrows
        if row_offset:
            self.current_row = row_offset
        else:
            self.current_row = 0

    def __iter__(self):
        return self

    def next(self):
        if self.current_row >= self.row_count:
            raise StopIteration

        row = self.sheet.row(self.current_row)
        row = [self._cell_value(cell) for cell in row]
        self.current_row += 1
        return row

    def _cell_value(self, cell):
        """Convert Excel cell into value of a python type
        
        (from Swiss XlsReader.cell_to_python)"""

        # annoying need book argument for datemode
        # info on types: http://www.lexicon.net/sjmachin/xlrd.html#xlrd.Cell-class
        if cell.ctype == xlrd.XL_CELL_NUMBER:
            return float(cell.value)
        elif cell.ctype == xlrd.XL_CELL_DATE:
            # TODO: distinguish date and datetime
            args = xlrd.xldate_as_tuple(cell.value, self.workbook.datemode)
            try:
                return datetime.date(args[0], args[1], args[2])
            except Exception, inst:
                # print 'Error parsing excel date (%s): %s' % (args, inst)
                return None
        elif cell.ctype == xlrd.XL_CELL_BOOLEAN:
            return bool(cell.value)
        else:
            return cell.value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import threading
import time
import brewery.streams as streams

class PipeTestCase(unittest.TestCase):
    def setUp(self):
        self.processed_count = 0
        self.pipe = None
        self.sent_count = 0

    def send_sample(self, sample_size = 10):
        for i in range(0, sample_size):
            self.sent_count += 1
            self.pipe.put(i)

    def send_sample_limit_watch(self, sample_size = 10):
        for i in range(0, sample_size):
            if self.pipe.closed():
                break
            self.sent_count += 1
            time.sleep(0.01)
            self.pipe.put(i)

    def source_function(self):
        self.send_sample(1000)
        self.pipe.done_sending()

    def source_limit_function(self):
        self.send_sample_limit_watch(1000)
        self.pipe.done_sending()

    
    def target_function(self):
        self.processed_count = 0
        for value in self.pipe.rows():
            self.processed_count += 1

    def target_limit_function(self):
        self.processed_count = 0
        for value in self.pipe.rows():
            self.processed_count += 1
            if self.processed_count >= 20:
                break
        self.pipe.done_receiving()


    def test_put_get(self):
        self.pipe = streams.Pipe(buffer_size = 10)
        src = threading.Thread(target=self.source_function)
        target = threading.Thread(target=self.target_function)
        src.start()
        target.start()
        target.join()
        src.join()
        self.assertEqual(self.processed_count, 1000)

    def test_early_get_finish(self):
        self.pipe = streams.Pipe(buffer_size = 10)
        src = threading.Thread(target=self.source_function)
        target = threading.Thread(target=self.target_limit_function)
        src.start()
        target.start()
        target.join()
        src.join()
        self.assertEqual(self.processed_count, 20)
        # self.assertEqual(self.sent_count, 1000)

    def test_early_get_finish_watched(self):
        self.pipe = streams.Pipe(buffer_size = 10)
        src = threading.Thread(target=self.source_limit_function)
        target = threading.Thread(target=self.target_limit_function)
        src.start()
        target.start()
        target.join()
        self.assertEqual(self.processed_count, 20)
        self.assertLess(self.sent_count, 1000)

class Pipe2TestCase(unittest.TestCase):

    def setUp(self):
        self.pipe = streams.Pipe(100)

    def stest_put_one(self):
        self.pipe.put(1)
        self.pipe.done_sending()
        for row in self.pipe.rows():
            pass
        self.assertEqual(1, row)

    def test_pget_one(self):
        for i in range(1,100):
            self.pipe.put(i)
        self.pipe.done_sending()

        row = None

        for row in self.pipe.rows():
            break

        self.assertEqual(1, row)

    def producer(self, count = 100, stop = None):
        from random import random as _random
        from time import sleep as _sleep

        counter = 0
        while counter < count:
            self.pipe.put(counter)
            _sleep(_random() * 0.00001)
            counter = counter + 1
        self.pipe.done_sending()

    def consumer(self, count = None):
        self.consumed_count = 0
        for row in self.pipe.rows():
            self.consumed_count += 1
            if count and self.consumed_count >= count:
                break
        self.pipe.done_receiving()

    def test_sending(self):
        producer = threading.Thread(target = self.producer)
        consumer = threading.Thread(target = self.consumer)
        producer.start()
        consumer.start()
        producer.join()
        consumer.join()
        self.assertEqual(100, self.consumed_count)

        self.pipe = streams.Pipe(100)
        producer = threading.Thread(target = self.producer, kwargs = {"count": 200})
        consumer = threading.Thread(target = self.consumer)
        producer.start()
        consumer.start()
        producer.join()
        consumer.join()
        self.assertEqual(200, self.consumed_count)

        self.pipe = streams.Pipe(100)
        producer = threading.Thread(target = self.producer, kwargs = {"count": 150})
        consumer = threading.Thread(target = self.consumer)
        producer.start()
        consumer.start()
        producer.join()
        consumer.join()
        self.assertEqual(150, self.consumed_count)

    def test_receiving(self):
        self.pipe = streams.Pipe(100)
        producer = threading.Thread(target = self.producer, kwargs = {"count": 15})
        consumer = threading.Thread(target = self.consumer, kwargs = {"count": 5})
        producer.start()
        consumer.start()
        producer.join()
        consumer.join()
        self.assertEqual(5, self.consumed_count)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base
import re
import brewery
import brewery.ds as ds
from brewery.common import FieldError
import itertools

class FieldMapNode(base.Node):
    """Node renames input fields or drops them from the stream.
    """
    node_info = {
        "type": "field",
        "label" : "Field Map",
        "description" : "Rename or drop fields from the stream.",
        "attributes" : [
            {
                "name": "map_fields",
                "label": "Map fields",
                "description": "Dictionary of input to output field name."
            },
            {
                "name": "drop_fields",
                "label": "drop fields",
                "description": "List of fields to be dropped from the stream - incompatible with keep_fields."
            },
            {
                "name": "keep_fields",
                "label": "keep fields",
                "description": "List of fields to keep from the stream - incompatible with drop_fields."
            }
        ]
    }

    def __init__(self, map_fields = None, drop_fields = None, keep_fields=None):
        super(FieldMapNode, self).__init__()

        if drop_fields and keep_fields:
            raise FieldError('Invalid configuration of FieldMapNode: you cant specify both keep_fields and drop_fields.')

        if map_fields:
            self.mapped_fields = map_fields
        else:
            self.mapped_fields = {}

        if drop_fields:
            self.dropped_fields = set(drop_fields)
        else:
            self.dropped_fields = set([])

        if keep_fields:
            self.kept_fields = set(keep_fields)
        else:
            self.kept_fields = set([])
            
        self._output_fields = []
        
    def rename_field(self, source, target):
        """Change field name"""
        self.mapped_fields[source] = target
    
    def drop_field(self, field):
        """Do not pass field from source to target"""
        self.dropped_fields.add(field)

    @property
    def output_fields(self):
        return self._output_fields

    def initialize(self):
        self.map = brewery.FieldMap(rename=self.mapped_fields, drop=self.dropped_fields, keep=self.kept_fields)
        self._output_fields = self.map.map(self.input.fields)
        self.filter = self.map.row_filter(self.input.fields)

    def run(self):
        self.mapped_field_names = self.mapped_fields.keys()

        for row in self.input.rows():
            row = self.filter.filter(row)
            self.put(row)

class TextSubstituteNode(base.Node):
    """Substitute text in a field using regular expression."""
    
    node_info = {
        "type": "field",
        "label" : "Text Substitute",
        "description" : "Substitute text in a field using regular expression.",
        "attributes" : [
            {
                "name": "field",
                "label": "substituted field",
                "description": "Field containing a string or text value where substition will "
                               "be applied"
            },
            {
                "name": "derived_field",
                "label": "derived field",
                "description": "Field where substition result will be stored. If not set, then "
                               "original field will be replaced with new value."
            },
            {
                "name": "substitutions",
                "label": "substitutions",
                "description": "List of substitutions: each substition is a two-element tuple "
                               "(`pattern`, `replacement`) where `pattern` is a regular expression "
                               "that will be replaced using `replacement`"
            }
        ]
    }

    def __init__(self, field, derived_field = None):
        """Creates a node for text replacement.
        
        :Attributes:
            * `field`: field to be used for substitution (should contain a string)
            * `derived_field`: new field to be created after substitutions. If set to ``None`` then the
              source field will be replaced with new substituted value. Default is ``None`` - same field
              replacement.
        
        """
        super(TextSubstituteNode, self).__init__()

        self.field = field
        self.derived_field = derived_field
        self.substitutions = []
        
    def add_substitution(self, pattern, repl):
        """Add replacement rule for field.
        
        :Parameters:
            * `pattern` - regular expression to be searched
            * `replacement` - string to be used as replacement, default is empty string
        """

        self.substitutions.append( (re.compile(pattern), repl) )
    
    # FIXME: implement this
    # @property
    # def output_fields(self):
    #     pass
        
    def run(self):
        pipe = self.input

        if self.derived_field:
            append = True
        else:
            append = False

        index = self.input_fields.index(self.field)
            
        for row in pipe.rows():
            value = row[index]
            for (pattern, repl) in self.substitutions:
                value = re.sub(pattern, repl, value)
            if append:
                row.append(value)
            else:
                row[index] = value

            self.put(row)


class StringStripNode(base.Node):
    """Strip spaces (orother specified characters) from string fields."""

    node_info = {
        "type": "field",
        "icon": "string_strip_node",
        "label" : "String Strip",
        "description" : "Strip characters.",
        "attributes" : [
            {
                "name": "fields",
                "description": "List of string fields to be stripped. If none specified, then all "
                               "fields of storage type `string` are stripped"
            },
            {
                "name": "chars",
                "description": "Characters to be stripped. "
                               "By default all white-space characters are stripped."
            }
        ]
    }

    def __init__(self, fields = None, chars = None):
        """Creates a node for string stripping.

        :Attributes:
            * `fields`: fields to be stripped
            * `chars`: characters to be stripped

        """
        super(StringStripNode, self).__init__()

        self.fields = fields
        self.chars = chars

    def run(self):

        if self.fields:
            fields = self.fields
        else:
            fields = []
            for field in self.input.fields:
                if field.storage_type == "string" or field.storage_type == "text":
                    fields.append(field)

        indexes = self.input_fields.indexes(fields)

        for row in self.input.rows():
            for index in indexes:
                value = row[index]
                if value:
                    row[index] = value.strip(self.chars)

            self.put(row)

class CoalesceValueToTypeNode(base.Node):
    """Coalesce values of selected fields, or fields of given type to match the type.
    
    * `string`, `text`
        * Strip strings
        * if non-string, then it is converted to a unicode string
        * Change empty strings to empty (null) values
    * `float`, `integer`
        * If value is of string type, perform string cleansing first and then convert them to
          respective numbers or to null on failure

    """

    node_info = {
        "type": "field",
        "icon": "coalesce_value_to_type_node",
        "description" : "Coalesce Value to Type",
        "attributes" : [
            {
                "name": "fields",
                "description": "List of fields to be cleansed. If none given then all fields "
                               "of known storage type are cleansed"
            },
            {
                "name": "types",
                "description": "List of field types to be coalesced (if no fields given)"
            },
            {
                "name": "empty_values",
                "description": "dictionary of type -> value pairs to be set when field is "
                               "considered empty (null)"
            }
        ]
    }

    def __init__(self, fields = None, types = None, empty_values = None):
        super(CoalesceValueToTypeNode, self).__init__()
        self.fields = fields
        self.types = types

        if empty_values:
            self.empty_values = empty_values
        else:
            self.empty_values = {}
        
    def initialize(self):
        if self.fields:
            fields = self.fields
        else:
            fields = self.input.fields

        self.string_fields = [f for f in fields if f.storage_type == "string"]
        self.integer_fields = [f for f in fields if f.storage_type == "integer"]
        self.float_fields = [f for f in fields if f.storage_type == "float"]
        
        self.string_indexes = self.input.fields.indexes(self.string_fields)
        self.integer_indexes = self.input.fields.indexes(self.integer_fields)
        self.float_indexes = self.input.fields.indexes(self.float_fields)
        
        self.string_none = self.empty_values.get("string")
        self.integer_none = self.empty_values.get("integer")
        self.float_none = self.empty_values.get("float")
        
    def run(self):
        
        for row in self.input.rows():
            for i in self.string_indexes:
                value = row[i]
                if type(value) == str or type(value) == unicode:
                    value = value.strip()
                elif value:
                    value = unicode(value)
                    
                if value == "" or value is None:
                    value = self.string_none

                row[i] = value

            for i in self.integer_indexes:
                value = row[i]
                if type(value) == str or type(value) == unicode:
                    value = re.sub(r"\s", "", value.strip())

                try:
                    value = int(value)
                except:
                    value = self.integer_none

                row[i] = value

            for i in self.float_indexes:
                value = row[i]
                if type(value) == str or type(value) == unicode:
                    value = re.sub(r"\s", "", value.strip())

                try:
                    value = float(value)
                except:
                    value = self.float_none

                row[i] = value
        
            self.put(row)

class ValueThresholdNode(base.Node):
    """Create a field that will refer to a value bin based on threshold(s). Values of `range` type
    can be compared against one or two thresholds to get low/high or low/medium/high value bins.

    *Note: this node is not yet implemented*
    
    The result is stored in a separate field that will be constructed from source field name and
    prefix/suffix.
    
    For example:
        * amount < 100 is low
        * 100 <= amount <= 1000 is medium
        * amount > 1000 is high

    Generated field will be `amount_threshold` and will contain one of three possible values:
    `low`, `medium`, `hight`
    
    Another possible use case might be for binning after data audit: we want to measure null 
    record count and we set thresholds:
        
        * ratio < 5% is ok
        * 5% <= ratio <= 15% is fair
        * ratio > 15% is bad
        
    We set thresholds as ``(0.05, 0.15)`` and values to ``("ok", "fair", "bad")``
        
    """
    
    node_info = {
        "type": "field",
        "label" : "Value Threshold",
        "description" : "Bin values based on a threshold.",
        "attributes" : [
            {
                "name": "thresholds",
                "description": "List of fields of `range` type and threshold tuples "
                               "(field, low, high) or (field, low)"
            },
            {
                "name": "bin_names",
                "description": "Names of bins based on threshold. Default is low, medium, high"
            },
            {
                "name": "prefix",
                "description": "field prefix to be used, default is none."
            },
            {
                "name": "suffix",
                "description": "field suffix to be used, default is '_bin'"
            }
        ]
    }

    def __init__(self, thresholds = None, bin_names = None, prefix = None, suffix = None):
        self.thresholds = thresholds
        self.bin_names = bin_names
        self.prefix = prefix
        self.suffix = suffix
        self._output_fields = None
    
    @property
    def output_fields(self):
        return self._output_fields
    
    def initialize(self):
        field_names = [t[0] for t in self.thresholds]

        self._output_fields = brewery.FieldList()

        for field in self.input.fields:
            self._output_fields.append(field)

        if self.prefix:
            prefix = self.prefix
        else:
            prefix = ""
            
        if self.suffix:
            suffix = self.suffix
        else:
            suffix = "_bin"

        for name in field_names:
            field = brewery.Field(prefix + name + suffix)
            field.storage_type = "string"
            field.analytical_type = "set"
            self._output_fields.append(field)

        input_fields = self.input.fields

        # Check input fields
        for name in field_names:
            if not name in self.input.fields:
                raise FieldError("No input field with name %s" % name)
                
        self.threshold_field_indexes = self.input.fields.indexes(field_names)
        
    def run(self):
        thresholds = []
        for t in self.thresholds:
            if len(t) == 1:
                # We have only field name, then use default threshold: 0
                thresholds.append( (0, ) )
            elif len(t) == 2:
                thresholds.append( (t[1], ) )
            elif len(t) >= 2:
                thresholds.append( (t[1], t[2]) )
            elif len(t) == 0:
                raise ValueError("Invalid threshold specification: should be field name, low and optional high")
        
        if not self.bin_names:
            bin_names = ("low", "medium", "high")
        else:
            bin_names = self.bin_names
        
        for row in self.input.rows():
            for i, t in enumerate(thresholds):
                value = row[self.threshold_field_indexes[i]]
                bin = None
                if len(t) == 1:
                    if value < t[0]:
                        bin = bin_names[0]
                    else:
                        bin = bin_names[-1]
                elif len(t) > 1:
                    if value < t[0]:
                        bin = bin_names[0]
                    if value > t[1]:
                        bin = bin_names[-1]
                    else:
                        bin = bin_names[1]

                row.append(bin)
            self.put(row)

class DeriveNode(base.Node):
    """Dreive a new field from other fields using an expression or callable function.

    The parameter names of the callable function should reflect names of the fields:

    .. code-block:: python

        def get_half(i, **args):
            return i / 2

        node.formula = get_half

    You can use ``**record`` to catch all or rest of the fields as dictionary:

    .. code-block:: python

        def get_half(**record):
            return record["i"] / 2
            
        node.formula = get_half
        

    The formula can be also a string with python expression where local variables are record field
    values:

    .. code-block:: python

        node.formula = "i / 2"

    """

    node_info = {
        "label" : "Derive Node",
        "description" : "Derive a new field using an expression.",
        "attributes" : [
            {
                 "name": "field_name",
                 "description": "Derived field name",
                 "default": "new_field"
            },
            {
                 "name": "formula",
                 "description": "Callable or a string with python expression that will evaluate to " \
                                "new field value"
            },
            {
                "name": "analytical_type",
                 "description": "Analytical type of the new field",
                 "default": "unknown"
            },
            {
                "name": "storage_type",
                 "description": "Storage type of the new field",
                 "default": "unknown"
            }
        ]
    }


    def __init__(self, formula = None, field_name = "new_field", analytical_type = "unknown",
                        storage_type = "unknown"):
        """Creates and initializes selection node
        """
        super(DeriveNode, self).__init__()
        self.formula = formula
        self.field_name = field_name
        self.analytical_type = analytical_type
        self.storage_type = storage_type
        self._output_fields = None

    @property
    def output_fields(self):
        return self._output_fields

    def initialize(self):
        if isinstance(self.formula, basestring):
            self._expression = compile(self.formula, "SelectNode condition", "eval")
            self._formula_callable = self._eval_expression
        else:
            self._formula_callable = self.formula

        self._output_fields = brewery.FieldList()

        for field in self.input.fields:
            self._output_fields.append(field)

        new_field = brewery.Field(self.field_name, analytical_type = self.analytical_type, 
                                  storage_type = self.storage_type)
        self._output_fields.append(new_field)

    def _eval_expression(self, **record):
        return eval(self._expression, None, record)

    def run(self):
        for record in self.input.records():
            if self._formula_callable:
                record[self.field_name] = self._formula_callable(**record)
            else:
                record[self.field_name] = None

            self.put_record(record)

class BinningNode(base.Node):
    """Derive a bin/category field from a value.

    .. warning::
    
        Not yet implemented
    
    Binning modes:
    
    * fixed width (for example: by 100)
    * fixed number of fixed-width bins
    * n-tiles by count or by sum
    * record rank
    
        
    """
    
    node_info = {
        "type": "field",
        "label" : "Binning",
        "icon": "histogram_node",
        "description" : "Derive a field based on binned values (histogram)"
    }
           